"""
    Run a scenario in a model.
    CRUD operations for model run options.

    Do not use this class directly, use FrogModel wrapper instead.
"""
from typing import Optional, Literal, TypedDict, get_args, Dict
from collections import defaultdict
import uuid
import optilogic
import pandas as pd

class RunScenarioResponse(TypedDict):
    """"
    Response from running a scenario
    """
    status: Literal["error", "success"]
    message: str


UI_DISPLAY_CATEGORIES = Literal[
    "Basic",
    "Advanced"
]

STATUS = Literal[
    "Include",
    "Exclude"
]

ENGINES = Literal[
    "neo",
    "throg",
    "triad",
    "dendro",
    "hopper",
    "dart"
]

RESOURCE_SIZES = Literal[
    'mini',
    '4xs',
    '3xs',
    '2xs',
    'xs',
    's',
    'm',
    'l',
    'xl',
    '2xl',
    '3xl',
    '4xl',
    'overkill',
]


class ModelRunOption(TypedDict):
    """
    Model Run Option
    """
    option: str
    status: Optional[STATUS] = "Include"
    value: str = ""
    technology: Optional[ENGINES] = "neo"
    description: str = ""
    datatype: str = "String"
    uidisplayname: str = ""
    uidisplaycategory: UI_DISPLAY_CATEGORIES = "Basic"
    uidisplayorder: Optional[int]
    uidisplaysubcategory: Optional[str]


class RunScenario:
    """
    Run a scenario in a model.
    CRUD operations for model run options.

    Do not use this class directly, use FrogModel wrapper instead.
    """
    def __init__(self, frog_model = None):
        if not frog_model:
            raise ValueError("Frog model is required.")

        self.frog_model = frog_model
        self.model_name = frog_model.model_name
        self.app_key = frog_model._app_key
        self.log = frog_model.log

        if self.app_key:
            self.api = optilogic.pioneer.Api(auth_legacy=False, appkey=self.app_key)
        else:
            raise ValueError("Could not authenticate, app_key is required to run a scenario")


    def __handle_error__(self, correlation_id, message):
        """
        Log an error and return a response
        """
        self.log.error("%s %s", correlation_id, message)
        raise ValueError(message)


    def get_all_run_parameters(self, engine: Optional[str] = '', correlation_id = None) -> dict:
        """
        Get all run parameters for a model
        """
        try:
            if not correlation_id:
                correlation_id = str(uuid.uuid4())

            self.log.info(
                "%s Getting all run parameters for model %s with engine %s",
                correlation_id,
                self.model_name,
                engine
            )

            sql_statement = "SELECT * FROM ModelRunOptions"
            params = {}
            if engine:
                if engine not in get_args(ENGINES):
                    return self.__handle_error__(correlation_id, f"Invalid engine {engine}, list of valid engines: {get_args(ENGINES)}")

                if engine == "neo_with_infeasibility":
                    engine = "neo"
                    self.log.info(
                        "%s Running with infeasibility check enabled, setting engine to neo",
                        correlation_id
                    )


                sql_statement += " WHERE technology ILIKE %(engine)s"
                params["engine"] = f"%{engine}%"

            self.log.info(
                "%s Running SQL statement: %s with params %s",
                correlation_id,
                sql_statement,
                params
            )

            data = self.frog_model.read_sql(sql_statement, params=params)

            if not data.empty:
                self.log.info(
                    "%s Run parameters fetched successfully",
                    correlation_id
                )
                return {
                    "status": "success",
                    "message": "Run parameters fetched successfully",
                    "data": data.to_dict(orient="records")
                }

            return self.__handle_error__(correlation_id, "Error fetching run parameters")
        except Exception as e:
            return self.__handle_error__(correlation_id, f"Error fetching run parameters: {str(e)}")


    def update_run_parameter_value(self, option: str, value: str, correlation_id = None) -> dict:
        """
        Update a run parameter value in a model
        """
        try:
            if not correlation_id:
                correlation_id = str(uuid.uuid4())

            self.log.info(
                "%s Updating run parameter %s in model %s with value %s",
                correlation_id,
                option,
                self.model_name,
                value
            )

            model_run_option = self.frog_model.read_sql(
                "SELECT * FROM ModelRunOptions WHERE LOWER(option) = %(option)s;",
                params={"option": option.lower()}
            )

            if model_run_option.empty:
                return self.__handle_error__(correlation_id, f"Run parameter {option} does not exist")

            self.log.info(
                "%s Run parameter %s exists in model %s",
                correlation_id,
                option,
                self.model_name
            )

            parameter_datatype = model_run_option.iloc[0]['datatype']

            self.log.info(
                "%s Run parameter %s has datatype %s checking if value is valid...",
                correlation_id,
                option,
                parameter_datatype
            )

            datatype_validation = self.__model_run_option_datatype_valid__(parameter_datatype, value, correlation_id)
            if datatype_validation["status"] == "error":
                return datatype_validation

            value = str(value)
            self.log.info(
                "%s Value is valid. Updating run parameter %s with value %s",
                correlation_id,
                option,
                value
            )

            self.frog_model.exec_sql(
                f"UPDATE ModelRunOptions SET value = '{value}' WHERE LOWER(option) = '{option.lower()}';"
            )

            self.log.info(
                "%s Run parameter %s updated successfully",
                correlation_id,
                option
            )

            return { "status": "success", "message": f"Run parameter {option} updated successfully" }
        except Exception as e:
            return self.__handle_error__(correlation_id, f"Error updating run parameter {option}: {str(e)}")


    def add_run_parameter(self, model_run_option: ModelRunOption, correlation_id = None) -> dict:
        """"
        Add a run parameter to a model

        model_name: str - The name of the model
        model_run_option: ModelRunOption - The run parameter to add
            option: str - The name of the run parameter
            status: str - The status of the run parameter, default: Include. Options: Include, Exclude. If Exclude, the run parameter will not be used in the scenario run and not be shown in the UI
            value: str - The value of the run parameter, default: "", has to follow the datatype rules
            technology: str - The technology of the run parameter, default: neo. Options: neo, neo_with_infeasibility, throg, triad, dendro, hopper, dart
            description: str - The description of the run parameter, default: ""
            datatype: str - The datatype of the run parameter, default: String. Options: String, int, double, [True, False], [custom1, custom2]
            uidisplaycategory: str - The UI display category of the run parameter, default: Basic. Options: Basic, Advanced. Basic means it will be shown in the CF UI
            uidisplayname: str - The CF UI display name of the run parameter, default: option. If not set, it will be the same as the option if not provided
            uidisplayorder: int - The order of the run parameter in the CF UI, default: highest order + 1, cannot be the same as another run parameter
        correlation_id: str - The correlation id for logging
        """
        try:
            if not correlation_id:
                correlation_id = str(uuid.uuid4())

            default_model_run_option = ModelRunOption(
                option="",
                status="Include",
                value="",
                technology="neo",
                description="",
                datatype="String",
                uidisplaycategory="Basic",
                uidisplayname="",
                uidisplayorder=None,
                uidisplaysubcategory=""
            )

            model_run_option = {**default_model_run_option, **model_run_option}

            ### Validate MRO parameters ###
            if not model_run_option["option"]:
                return self.__handle_error__(correlation_id, "Option parameter is required")

            if model_run_option["status"] not in get_args(STATUS):
                return self.__handle_error__(correlation_id, f"Invalid status {model_run_option['status']}, list of valid statuses: {get_args(STATUS)}")

            if model_run_option["technology"] == "neo_with_infeasibility":
                model_run_option["technology"] = "neo"
                self.log.info(
                    "%s Setting technology to neo",
                    correlation_id
                )

            if model_run_option["technology"] not in get_args(ENGINES):
                return self.__handle_error__(correlation_id, f"Invalid technology {model_run_option['technology']}, list of valid technologies: {get_args(ENGINES)}")

            if not model_run_option["uidisplaycategory"] in get_args(UI_DISPLAY_CATEGORIES):
                return self.__handle_error__(correlation_id, f"Invalid UI display category {model_run_option['uidisplaycategory']}, list of valid categories: {get_args(UI_DISPLAY_CATEGORIES)}")

            datatype_validation = self.__model_run_option_datatype_valid__(model_run_option["datatype"], model_run_option["value"], correlation_id)

            if datatype_validation["status"] == "error":
                return datatype_validation

            all_model_run_options = self.get_all_run_parameters(correlation_id=correlation_id)

            # check if MRO with column option already exists
            for option in all_model_run_options["data"]:
                if option["option"].lower() == model_run_option["option"].lower():
                    return self.__handle_error__(correlation_id, f"Run parameter {model_run_option['option']} already exists")

            if not model_run_option["uidisplayname"]:
                model_run_option["uidisplayname"] = model_run_option["option"]
                self.log.info(
                    "%s Setting uidisplayname to %s",
                    correlation_id,
                    model_run_option["option"]
                )

            # If uidisplayorder is not set, get the highest order and increment
            # if is set check if it already exists
            if model_run_option["uidisplayorder"] is None:
                self.log.info(
                    "%s Getting highest order for uidisplayorder",
                    correlation_id
                )

                highest_order = 0
                for option in all_model_run_options["data"]:
                    display_order = int(option["uidisplayorder"])
                    highest_order = max(highest_order, display_order)

                model_run_option["uidisplayorder"] = highest_order + 1
                self.log.info(
                    "%s Setting uidisplayorder to %s",
                    correlation_id,
                    model_run_option["uidisplayorder"]
                )
            else:
                for option in all_model_run_options["data"]:
                    if option["uidisplayorder"] == str(model_run_option["uidisplayorder"]):
                        return self.__handle_error__(correlation_id, f"Run parameter with order {model_run_option['uidisplayorder']} already exists")

            model_run_option["technology"] = f"[{model_run_option['technology'].upper()}]"
            self.log.info(
                "%s Setting technology to %s",
                correlation_id,
                model_run_option["technology"]
            )

            self.frog_model.exec_sql(
                f"INSERT INTO ModelRunOptions (option, status, value, technology, description, datatype, uidisplaycategory, uidisplayname, uidisplayorder) VALUES ('{model_run_option['option']}', '{model_run_option['status']}', '{model_run_option['value']}', '{model_run_option['technology']}', '{model_run_option['description']}', '{model_run_option['datatype']}', '{model_run_option['uidisplaycategory']}', '{model_run_option['uidisplayname']}', {model_run_option['uidisplayorder']});"
            )

            self.log.info(
                "%s Run parameter %s added successfully",
                correlation_id,
                model_run_option["option"]
            )

            return { "status": "success", "message": f"Run parameter {model_run_option['option']} added successfully", "model_run_option": model_run_option }
        except Exception as e:
            return self.__handle_error__(correlation_id, f"Error adding run parameter {model_run_option['option']}: {str(e)}, {model_run_option}")


    def delete_run_parameter(self, option: str, correlation_id = None) -> dict:
        """
        Delete a model run option

        Args:
            option: str -- The name of the run parameter

        Returns:
            dict: A dictionary containing the status and message
        """
        try:
            if not correlation_id:
                correlation_id = str(uuid.uuid4())

            self.log.info(
                "%s Deleting run parameter %s in model %s",
                correlation_id,
                option,
                self.model_name
            )

            model_run_option = self.frog_model.read_sql(
                "SELECT * FROM ModelRunOptions WHERE LOWER(option) = %(option)s;",
                params={"option": option.lower()}
            )

            if model_run_option.empty:
                return self.__handle_error__(correlation_id, f"Run parameter {option} does not exist")

            self.log.info(
                "%s Run parameter %s exists in model %s",
                correlation_id,
                option,
                self.model_name
            )

            self.frog_model.exec_sql(
                f"DELETE FROM ModelRunOptions WHERE LOWER(option) = '{option.lower()}';"
            )

            self.log.info(
                "%s Run parameter %s deleted successfully",
                correlation_id,
                option
            )

            return { "status": "success", "message": f"Run parameter {option} deleted successfully" }
        except Exception as e:
            return self.__handle_error__(correlation_id, f"Error deleting run parameter {option}: {str(e)}")


    def get_job_logs(self, job_key: str, workspace: str = "Studio", correlation_id = None) -> Dict[str, str]:
        """
        Get the logs of a job

        Args:
            job_key: str -- The key of the job
            workspace: str -- The workspace of the job
            correlation_id: str -- The correlation id for logging
        
        Returns:
            dict: A dictionary containing the status, message, and logs
        """
        if not correlation_id:
            correlation_id = str(uuid.uuid4())

        if not workspace:
            self.__handle_error__(correlation_id, "Workspace is required.")

        if not job_key:
            self.__handle_error__(correlation_id, "Job key is required.")

        try:
            result = self.api.wksp_job_file_result(jobkey=job_key, wksp=workspace)

            if not isinstance(result, str) and result.get("result") == "error":
                return self.__handle_error__(correlation_id, f"Error fetching logs for job {job_key}: {result.get('message')}")

            return {"status": "success", "message": "", "logs": result}
        except Exception as e:
            return self.__handle_error__(correlation_id, f"Error fetching logs for job {job_key}: {str(e)}")


    # TODO:
    # filter by technology
    # paginate
    # less details
    # is this good logic with iteration if big data?
    def all_scenarios_preview(self, correlation_id = None) -> dict:
        """
        Get all scenarios preview

        Args:
            correlation_id: str -- The correlation id for logging
        
        Returns:
            dict: A dictionary containing the runnable scenarios, their items, and the technology
        """

        query = """
            SELECT
                s.scenarioname AS scenario_name,
                s.technology AS technology,
                s.notes AS scenario_notes,
                si.itemname AS item_name,
                si.tablename AS table_name,
                si.action AS action,
                si.condition AS condition,
                si.conditiontype AS condition_type,
                si.namedfilters AS named_filters,
                si.notes AS item_notes,
                sia.itemorder AS item_order
            FROM
                scenarios s
            LEFT JOIN
                scenarioitemassignments sia
                ON s.scenarioname = sia.scenarioname
            LEFT JOIN
                scenarioitems si
                ON sia.itemname = si.itemname
            ORDER BY
                s.scenarioname, sia.itemorder;
        """

        try:
            if not correlation_id:
                correlation_id = str(uuid.uuid4())

            self.log.info(
                "%s Getting all scenarios for model %s",
                correlation_id,
                self.model_name
            )

            data = self.frog_model.read_sql(query)

            if data.empty:
                return { "status": "success", "data": [] }

            scenarios_dict = defaultdict(lambda: {"scenarioname": "", "technology": "", "notes": "", "scenarioitems": []})

            for _, row in data.iterrows():
                scenario_name = row["scenario_name"]

                # If the scenario doesn't exist, initialize it
                if not scenarios_dict[scenario_name]["scenarioname"]:
                    scenarios_dict[scenario_name]["scenarioname"] = scenario_name
                    scenarios_dict[scenario_name]["technology"] = row["technology"]
                    scenarios_dict[scenario_name]["notes"] = row["scenario_notes"]

                # If there are scenario items, add them to the list
                if pd.notna(row["item_name"]):
                    scenarios_dict[scenario_name]["scenarioitems"].append({
                        "itemname": row["item_name"],
                        "tablename": row["table_name"],
                        "action": row["action"],
                        "condition": row["condition"],
                        "conditiontype": row["condition_type"],
                        "namedfilters": row["named_filters"],
                        "notes": row["item_notes"],
                        "itemorder": row["item_order"]
                    })

            # Convert the defaultdict to a list of scenarios
            scenarios = list(scenarios_dict.values())

            # Sort scenario items by item_order
            for scenario in scenarios:
                scenario["scenarioitems"].sort(key=lambda x: x["itemorder"] if x.get("itemorder") is not None else float("inf"))

            self.log.info(
                "%s Scenarios preview fetched successfully",
                correlation_id
            )

            return {
                "status": "success",
                "message": "Scenarios preview fetched successfully",
                "data": scenarios
            }
        except Exception as e:
            return self.__handle_error__(correlation_id, f"Error fetching scenarios: {str(e)}")


    def __neo_infeasibility_check_procedure__(
        self,
        run_neo_with_infeasibility: bool = False,
        correlation_id: str = None
    ):
        """
        Run NEO infeasibility check procedure.
        Has to update ModelRunOptions to False
        if not running infeasibility if the last check was set to True.

        Args:
            run_neo_with_infeasibility: bool -- Run the scenario with infeasibility check enabled - Default False
            correlation_id: str -- Correlation id for logging / tracing (for internal use)
        
        Returns:
            None or str: The engine to run the scenario or None
        """
        infeasibility_update_value = str(run_neo_with_infeasibility)

        self.log.info(
            "%s Updating infeasibility check value to %s",
            correlation_id,
            infeasibility_update_value
        )

        # TODO: update or create?? There is an option that this parameter does not exist
        self.frog_model.exec_sql(
            f"UPDATE ModelRunOptions SET value = '{infeasibility_update_value}' WHERE option = 'CheckInfeasibility';"
        )


    def __check_if_contains_unkown_scenarios__(self, scenarios: [str], correlation_id: str) -> Dict[str, str]:
        """"        
        Check if all of the scenario names exist in the model.

        If one does not exist, return an error message.

        Args:
            scenarios: [str] -- Name of the scenario or list of scenarios to check
            correlation_id: str -- Correlation id for logging / tracing (for internal use)
        
        Returns:
            dict: A dictionary containing the status and message
        """

        try:
            self.log.info(
                "%s Checking if scenario names exist in a model: %s",
                correlation_id,
                scenarios
            )

            if not scenarios:
                self.log.info(
                    "%s No scenarios to check in DB after filtering out baseline/all",
                    correlation_id
                )

                return {}

            query = f"""
                SELECT
                    ScenarioName,
                    CASE 
                        WHEN technology IS NULL OR TRIM(technology) = '' THEN 'NEO'
                        ELSE technology
                    END AS technology
                FROM Scenarios
                WHERE ScenarioName IN ('{"', '".join(scenarios)}');
            """

            result_df = self.frog_model.read_sql(
                query
            )

            existing_scenarios = set(result_df['scenarioname'].unique())
            input_scenarios = set(scenarios)

            missing_scenarios = input_scenarios - existing_scenarios

            if missing_scenarios:
                self.__handle_error__(
                    correlation_id,
                    f"Scenarios {missing_scenarios} do not exist. "
                    "Make sure the scenario names are correct and remember names are case sensitive."
                )

            self.log.info(
                "%s Scenarios %s exist in model %s",
                correlation_id,
                scenarios,
                self.model_name
            )

            return result_df.groupby('technology')['scenarioname'].apply(list).to_dict()
        except Exception as e:
            return {
                "status": True,
                "message": self.__handle_error__(
                    correlation_id,
                    f"Error checking if scenario names exist in a model: {str(e)}"
                )
            }


    def __run_scenario_tags__(self, engine: ENGINES, scenarios: [str], tags: str) -> str:
        final_tags = f"ModelName={self.model_name}, Engine={engine}, Scenarios={','.join(scenarios)}"
        if tags:
            final_tags += f", {tags}"

        return final_tags


    def __build_command_args_for_run__(
        self,
        scenario_list_for_engine: [str],
        engine: ENGINES,
        resource_size: RESOURCE_SIZES,
        wksp: str,
        version: str
    ) -> str:
        command_args_example = (
            f"-m '{self.model_name}' "
            f"-s '{','.join(scenario_list_for_engine)}' "
            f"-e '{engine.lower()}' "
            f"-r '{resource_size}' "
            f"-w '{wksp}'"
        )

        if version:
            command_args_example += f" -v '{version}'"

        return command_args_example


    def __determine_engines_and_scenarios_to_run__(
        self,
        scenarios: [str],
        engine: ENGINES,
        correlation_id: str
    ) -> [ENGINES, dict]:
        """
        Determine engines and scenarios to run based on the input.

        If "all" is passed then check if user passed engine or not.
        If engine is passed, group all scenarios under the engine.
        If engine is not passed, group all scenarios under the engine from the scenario table.
        If multiple engines are found, fetch and group scenarios by technology.

        If 'all' is not passed, check if all scenarios exist in the model.
        If not, return an error message.
        If all scenarios exist, group them by engine.
        If engine is passed, group all scenarios under the engine.
        If engine is not passed, group all scenarios under the engine from the scenario table.

        Then convert engine to list if it is a string.

        Validate if the engine is in the list of valid engines.

        Args:
            scenarios: [str] -- Name of the scenario or list of scenarios to run
            engine: ENGINES -- Engine to run the scenario on
            correlation_id: str -- Correlation id for logging / tracing (for internal use)
        
        Returns:
            [ENGINES, dict]: The engine to run the scenario on and the grouped scenarios
        """
        grouped_scenarios = {}

        if "all" in scenarios[0].lower():
            if engine:
                self.log.info(
                    "%s Scenario name 'all' passed, engine passed as well, setting engine to %s",
                    correlation_id
                )

                grouped_scenarios = {engine: 'All'}
            else:
                self.log.info(
                    "%s Scenario name 'all' passed, but without engine. Checking if there are more than one technology in the model",
                    correlation_id
                )

                query = """
                    SELECT DISTINCT 
                        CASE 
                            WHEN technology IS NULL OR TRIM(technology) = '' THEN 'NEO'
                            ELSE technology
                        END AS technology
                    FROM Scenarios;
                """

                all_technologies_df = self.frog_model.read_sql(query)
                existing_technologies = set(all_technologies_df['technology'])

                if len(existing_technologies) == 0:
                    self.log.info(
                        "%s No technology found in the model, defaulting to NEO",
                        correlation_id
                    )

                    grouped_scenarios = {"neo": 'All'}
                    engine = "neo"

                if len(existing_technologies) == 1:
                    self.log.info(
                        "%s Only one technology found in the model, setting engine to %s",
                        correlation_id,
                        existing_technologies
                    )

                    grouped_scenarios = {all_technologies_df['technology'][0]: 'All'}
                    engine = all_technologies_df['technology'][0]

                if len(existing_technologies) > 1:
                    self.log.info(
                        "%s More than one technology found in the model, fetching all scenarios and grouping by technology",
                        correlation_id
                    )

                    query = """
                        SELECT ScenarioName, 
                            CASE 
                                WHEN technology IS NULL OR TRIM(technology) = '' THEN 'NEO'
                                ELSE technology
                            END AS technology
                        FROM Scenarios;
                    """

                    grouped_scenarios = self.frog_model.read_sql(query).groupby('technology')['scenarioname'].apply(list).to_dict()
                    engine = list(grouped_scenarios.keys())
        else:
            # Filter out "baseline" from the names to explicitly check
            names_to_check = [name for name in scenarios if name.lower() not in ["baseline"]]

            # Check if all of the scenario names exist in the model
            contains_unkown_scenarios_response = self.__check_if_contains_unkown_scenarios__(
                names_to_check,
                correlation_id
            )

            # if engine passed, group all scenarios under the engine
            # if engine not passed, group all scenarios under the engine from the scenario table
            if engine:
                grouped_scenarios = {engine: scenarios}
            else:
                grouped_scenarios = contains_unkown_scenarios_response
                self.log.info(
                    "%s No engine passed, fetching engines from the model",
                    correlation_id
                )
                engines = list(grouped_scenarios.keys())
                self.log.info(
                    "%s Engines found in the model: %s",
                    correlation_id,
                    grouped_scenarios
                )

                if len(engines) == 0:
                    self.log.info(
                        "%s No engines found in the model, defaulting to NEO",
                        correlation_id
                    )
                    engine = "neo"
                    
                    if len(scenarios) > 0:
                        grouped_scenarios = {"neo": scenarios}
                    else:
                        grouped_scenarios = {"neo": "Baseline"} # TODO ?

                if len(engines) == 1:
                    engine = engines[0]
                    self.log.info(
                        "%s One engine found in the model, setting engine to %s",
                        correlation_id,
                        engine
                    )

                if len(engines) > 1:
                    self.log.info(
                        "%s More than one engine found in the model, setting engine to list of engines",
                        correlation_id
                    )
                    engine = engines

        if isinstance(engine, str):
            engine = [engine]

        # Check if valid engines in list ENGINES
        for engine_i in engine:
            if engine_i.lower() not in get_args(ENGINES):
                return self.__handle_error__(
                    correlation_id,
                    f"Invalid engine {engine_i}, list of valid engines: {get_args(ENGINES)}"
                )

        return engine, grouped_scenarios


    def __validate_run_parameters__(
        self,
        scenarios: [str],
        resource_size: RESOURCE_SIZES,
        engine: ENGINES,
        correlation_id: str
    ):
        """
        Validate run parameters before running the scenario.

        Args:
            scenarios: [str] -- Name of the scenario or list of scenarios to run
            resource_size: RESOURCE_SIZES -- Resource size to run the scenario on
            engine: ENGINES -- Engine to run the scenario on
            correlation_id: str -- Correlation id for logging / tracing (for internal use)
        """

        if not isinstance(scenarios, list):
            return self.__handle_error__(correlation_id, "Scenarios must be passed as a list. Example: ['Scenario']")

        if not scenarios or len(scenarios) == 0:
            return self.__handle_error__(correlation_id, "Must pass at least one scenario to run.")

        if resource_size not in get_args(RESOURCE_SIZES):
            return self.__handle_error__(
                correlation_id,
                f"Invalid resource size {resource_size}, valid: {get_args(RESOURCE_SIZES)}"
            )

        if engine and engine not in get_args(ENGINES):
            return self.__handle_error__(
                correlation_id,
                f"Invalid engine {engine}, list of valid engines: {get_args(ENGINES)}."
            )

        if "all" in [name.lower() for name in scenarios] and len(scenarios) > 1:
            return self.__handle_error__(
                correlation_id,
                "Cannot run all scenarios and specific scenarios at the same time. "
                "If you want to run all scenarios, pass ['All'] only."
            )

    # TODO:
    # IF user says Baseline without technology?
    # What if Baseline passed but scenario does not exist in the DB?
    # Test if model has no scenarios, but passed "All" or "Baseline"
    # scenarios to run with filter sql condition as parameter
    # debug, extra logs option run... model run option specific update/create ... add as parameter to function
    def run(
        self,
        scenarios: list[str] = ["Baseline"],
        wksp: str = "Studio",
        engine: None | ENGINES = None,
        run_neo_with_infeasibility: bool = False,
        resource_size: RESOURCE_SIZES = "s",
        tags: str = "",
        version: str = "",
        fire_and_forget: bool = False,
        correlation_id: str | None = None,
        check_configuration_before_run: bool = False
    ) -> RunScenarioResponse:
        """
        Run scenarios within a model.
        Can run all scenarios in a model by passing ["All"].
        If scenario names are not passed, will run the Baseline scenario (default).
        Scenario names are case sensitive.

        If engine is not passed, will run on the default
        engine from the scenario table (technology column).
        If technology is not set in the scenario table, will default to "neo".
        If technology is set to unknown value, will return an error.

        By default function will run the scenario on resource size S, which can be
        changed by passing resource_size parameter.

        By default function will await all of the scenarios to complete. Pass
        `fire_and_forget=True` to run in the background.

        If not sure about configuration before running, pass
        `check_configuration_before_run=True` to return the final config
        without actually running anything.

        If running with infeasibility check, set run_neo_with_infeasibility=True;
        engine parameter will be ignored and set to "neo".

        Args:
            scenarios (list[str]): Name(s) of the scenario(s) to run. Default ["Baseline"]
            wksp (str): Workspace to run the scenario in - Default "Studio"
            engine (str | list[str] | None): Engine(s) to run on - Default is derived from scenario table
            run_neo_with_infeasibility (bool): If True, forcibly run "neo" with infeasibility check
            resource_size (str): Resource size to run the scenario on - default "s"
            tags (str): Tags to run the scenario with
            version (str): Version of the sovler to run
            fire_and_forget (bool): If True, do not monitor the job. Default False
            correlation_id (str | None): Correlation ID for logs/tracing. If None, auto-generated
            check_configuration_before_run (bool): If True, only return the config, do not run anything.

        Returns:
            RunScenarioResponse: a dict describing success/failure status, job keys, etc.
        """
        correlation_id = correlation_id or str(uuid.uuid4())

        try:
            self.log.info(
                "%s Starting process to run scenario(s) %s in model %s",
                correlation_id,
                scenarios,
                self.model_name
            )

            # 1) Set engine to NEO if infeasibility check is enabled
            if run_neo_with_infeasibility:
                if engine and engine != "neo": # Inform user that engine will be set to neo
                    self.log.info(
                        "%s Running with infeasibility check enabled, setting engine to neo",
                        correlation_id
                    )

                engine = "neo"

            # 2) Validate run parameters
            self.__validate_run_parameters__(scenarios, resource_size, engine, correlation_id)

            # 3) Determine engine and scenarios to run
            engine, grouped_scenarios = self.__determine_engines_and_scenarios_to_run__(
                scenarios, engine, correlation_id
            )

            # 4) Run NEO infeasibility check procedure
            self.__neo_infeasibility_check_procedure__(run_neo_with_infeasibility, correlation_id)

            # 5) Check and return configuration before run if enabled
            if check_configuration_before_run:
                config_objects = []

                for e in engine:
                    scenario_list_for_engine = grouped_scenarios.get(e, [])

                    config_objects.append({
                        "command_args_example": self.__build_command_args_for_run__(scenario_list_for_engine, e, resource_size, wksp, version),
                        "tags": self.__run_scenario_tags__(e, scenario_list_for_engine, tags)
                    })

                return {
                    "status": "success",
                    "message": "Configuration check before run is enabled, no scenario run will be executed. Configuration check passed successfully. To run the scenario, set check_configuration_before_run to False",
                    "model_name": self.model_name,
                    "wksp": wksp,
                    "configurations": config_objects
                }

            # 6) Kick off Solvers
            job_key_by_engine = {}
            for e in engine:
                scenario_list_for_engine = grouped_scenarios.get(e, [])

                command_args = self.__build_command_args_for_run__(
                    scenario_list_for_engine,
                    e,
                    resource_size,
                    wksp,
                    version
                )

                self.log.info(
                    "%s Starting scenario(s) on engine '%s' with command_args: %s",
                    correlation_id,
                    e,
                    command_args
                )

                run_model_result = self.api.wksp_job_start(
                    wksp=wksp,
                    file_path="",
                    commandArgs=command_args,
                    tags=self.__run_scenario_tags__(e, scenario_list_for_engine, tags),
                    command="run_model",
                    resourceConfig='4xs'
                )

                if not run_model_result or 'jobKey' not in run_model_result:
                    return self.__handle_error__(
                        correlation_id,
                        f"Error starting scenario(s) {scenario_list_for_engine} on engine '{e}'"
                    )

                job_key = run_model_result['jobKey']
                job_key_by_engine[e] = job_key
                self.log.info(
                    "%s Scenario(s) %s started with job key %s on engine '%s'",
                    correlation_id,
                    scenario_list_for_engine,
                    job_key,
                    e
                )

            # 7) If fire and forget, return job keys
            if fire_and_forget:
                return {
                    "status": "success",
                    "message": "Scenario(s) started successfully (fire and forget).",
                    "job_keys": job_key_by_engine
                }

            # 8) Await scenarios to complete
            jobs_information_by_engine = {}
            for e, key in job_key_by_engine.items():
                self.log.info(
                    "%s Awaiting scenario(s) on engine '%s' with job key %s to complete",
                    correlation_id,
                    e,
                    key
                )
                is_done = self.api.util_job_monitor(
                    wksp=wksp,
                    job_key=key,
                    secs_max=86399,
                    stop_when="done"
                )
                if not is_done:
                    return self.__handle_error__(
                        correlation_id,
                        f"Scenario(s) on engine '{e}' did not complete within 24 hours"
                    )

                child_jobs_response = self.api.wksp_jobs(wksp=wksp, tags=f"ParentJob={key}")
                jobs = []
                for job in child_jobs_response['jobs']:
                    scenario_name = job['jobInfo']['tags'].split(", ")
                    scenario_name = [s.split("=")[1] for s in scenario_name if "Scenario=" in s][0]
                    jobs.append({
                        "job_key": job['jobKey'],
                        "status": job['status'],
                        "scenario_name": scenario_name
                    })

                jobs_information_by_engine[e] = {
                    "count_per_status": child_jobs_response['statusCounts'],
                    "jobs": jobs
                }

            return {
                "status": "success",
                "message": "Scenario(s) completed successfully.",
                "job_run_information": jobs_information_by_engine
            }
        except Exception as e:
            return { "status": "error", "message": f"Error running scenario(s): {str(e)}" }


    def __model_run_option_datatype_valid__(self, datatype: str, value: str, correlation_id: str) -> dict:
        """
        Check if the datatype of a model run option is valid
        """

        if datatype == "String":
            return { "status": "success", "message": "Value is a string" }

        if datatype == "int":
            try:
                int(value)
                return { "status": "success", "message": "Value is an integer" }
            except ValueError:
                return self.__handle_error__(correlation_id, "Value must be an integer, example: '1'")

        if datatype == "double":
            try:
                float(value)
                return { "status": "success", "message": "Value is a float" }
            except ValueError:
                return self.__handle_error__(correlation_id, "Value must be a float, example: '1.0'")

        if datatype == "[True, False]":
            if value.lower() in ["true", "false"]:
                return { "status": "success", "message": "Value is a boolean" }

            return self.__handle_error__(correlation_id, "Value must be True or False as type string, example: 'True' or 'False'")

        if datatype.startswith("[") and datatype.endswith("]"):
            custom_values = datatype[1:-1].split(", ")
            if value not in custom_values:
                return self.__handle_error__(correlation_id, f"Value must be one of {custom_values}")

            return { "status": "success", "message": f"Value is one of {custom_values}" }

        return self.__handle_error__(correlation_id, f"Unsupported datatype {datatype}, supported datatypes: String, int, double, [True, False], [custom1, custom2]")


    # def get_last_run_recommendation(self, scenario_name: Optional[str] = '', job_key: Optional[str]= '', workspace: str = "Studio") -> Dict[str, str]:
        # TODO: Some of the jobs have it, some don't
        # print("get_last_run_recommendation")

    # def get_last_run_resource_metrics(self, scenario_name: Optional[str] = '', job_key: Optional[str]= '', workspace: str = "Studio") -> Dict[str, str]:
    #     wksp_job_metrics_max

    ## Can track multiple scenarios by tags -> util_job_monitor_set

    # def create scenario - validation named filter per table, name validation, technology validation
    # def all named filters/per table
